if (not Grid) then return end

-- Libraries

local L = AceLibrary("AceLocale-2.2"):new("GridLayoutPlus")
local RL = AceLibrary("Roster-2.1")

-- New Grid module

GridLayoutPlus = Grid:NewModule("GridLayoutPlus", "AceConsole-2.0")

GridLayoutPlus.defaultDB = {
	autoLayout = true,
	layoutSorting = "By Group",
	showPets = true,
	showTanks = nil,
	showPetSpacer = nil,
	showTankSpacer = nil,
	unitsPerColumn = 5,
	classOrder = "WARRIOR,ROGUE,DRUID,PALADIN,SHAMAN,PRIEST,MAGE,WARLOCK,HUNTER",
	petOrder = "HUNTER,WARLOCK",
	roleOrder = "MAINTANK,MAINASSIST",
	maxGridColumns = 18,
	warnPetsHidden = nil,
	warnTanksHidden = nil,
}


-- Locals

local playerColumns = 0
local petColumns = 0
local tankColumns = 0
local tankType = nil
local raTankList = nil

local updateNeeded = nil


-- Default layout

GridLayoutPlus.DefaultLayout = {
	defaults = {
		maxColumns = 8,
	},
	[1] = nil,
	[2] = nil,
	[3] = nil,
	[4] = nil,
	[5] = nil,
	[6] = nil,
	[7] = nil,
	[8] = nil,
	[9] = nil,
	[10] = nil,
}

GridLayout:AddLayout(L["GridLayoutPlus"], GridLayoutPlus.DefaultLayout)


-- Init

function GridLayoutPlus:OnEnable()
	local options = GridLayout.options.args
	local ORDER_PLUS = 40
	options["PlusHeader"] = {
		type = "header",
		order = ORDER_PLUS,
	}
	ORDER_PLUS = ORDER_PLUS + 1
	options["autoLayout"] = {
		type = "toggle",
		name = L["Auto adjust layout (+)"],
		desc = L["Automatically adjust layout to fit raid size"],
		order = ORDER_PLUS,
		get = function ()
			return GridLayoutPlus.db.profile.autoLayout
		end,
		set = function (v)
			GridLayoutPlus.db.profile.autoLayout = v
			GridLayoutPlus:FullUpdate()
		end,
	}
	ORDER_PLUS = ORDER_PLUS + 1
	options["layoutSorting"] = {
		type = "text",
		name = L["Layout sorting (+)"],
		desc = L["Choose between By Group or By Class"],
		order = ORDER_PLUS + 1,
		get = function ()
			return GridLayoutPlus.db.profile.layoutSorting
		end,
		set = function (v)
			GridLayoutPlus.db.profile.layoutSorting = v
			updateNeeded = true
			GridLayoutPlus:RosterUpdate()
		end,
		validate = {
			["By Group"] = L["By Group"],
			["By Class"] = L["By Class"],
		},
	}
	ORDER_PLUS = ORDER_PLUS + 1
	options["showPets"] = {
		type = "toggle",
		name = L["Show pets (+)"],
		desc = L["Show pets when auto-adjusting (if possible)"],
		order = ORDER_PLUS,
		get = function ()
			return GridLayoutPlus.db.profile.showPets
		end,
		set = function (v)
			GridLayoutPlus.db.profile.showPets = v
			updateNeeded = true
			GridLayoutPlus:RosterUpdate()
		end,
	}
	ORDER_PLUS = ORDER_PLUS + 1
	options["showTanks"] = {
		type = "toggle",
		name = L["Show tanks (+)"],
		desc = L["Show tanks when auto-adjusting (if possible)"],
		order = ORDER_PLUS,
		get = function ()
			return GridLayoutPlus.db.profile.showTanks
		end,
		set = function (v)
			GridLayoutPlus.db.profile.showTanks = v
			GridLayoutPlus:FullUpdate()
		end,
	}
	ORDER_PLUS = ORDER_PLUS + 1
	options["showPetSpacer"] = {
		type = "toggle",
		name = L["Show pet spacer (+)"],
		desc = L["Show a spacer column between players and pets"],
		order = ORDER_PLUS,
		get = function ()
			return GridLayoutPlus.db.profile.showPetSpacer
		end,
		set = function (v)
			GridLayoutPlus.db.profile.showPetSpacer = v
			updateNeeded = true
			GridLayoutPlus:UpdateLayout()
		end,
	}
	ORDER_PLUS = ORDER_PLUS + 1
	options["showTankSpacer"] = {
		type = "toggle",
		name = L["Show tank spacer (+)"],
		desc = L["Show a spacer column between tanks and players"],
		order = ORDER_PLUS,
		get = function ()
			return GridLayoutPlus.db.profile.showTankSpacer
		end,
		set = function (v)
			GridLayoutPlus.db.profile.showTankSpacer = v
			updateNeeded = true
			GridLayoutPlus:UpdateLayout()
		end,
	}
	ORDER_PLUS = ORDER_PLUS + 1
	options["PlusAdvanced"] = {
		type = "group",
		name = L["Plus Advanced (+)"],
		desc = L["Advanced options for GridLayoutPlus"],
		order = ORDER_PLUS,
		args = {
			["unitsPerColumn"] = {
				type = "range",
				name = L["Units per column"],
				desc = L["How many units to show in a single column"],
				order = 1,
				min = 5,
				max = 40,
				step = 5,
				get = function()
					return GridLayoutPlus.db.profile.unitsPerColumn
				end,
				set = function(v)
					GridLayoutPlus.db.profile.unitsPerColumn = v
					updateNeeded = true
					GridLayoutPlus:FullUpdate()
				end,
			},
			["classOrder"] = {
				type = "text",
				name = L["Class order"],
				desc = L["Order in which player units appear"],
				usage = L["Order in which player units appear"],
				order = 2,
				get = function()
					return GridLayoutPlus.db.profile.classOrder
				end,
				set = function(v)
					GridLayoutPlus.db.profile.classOrder = v
					updateNeeded = true
					GridLayoutPlus:RosterUpdate()
				end,
			},
			["petOrder"] = {
				type = "text",
				name = L["Pet order"],
				desc = L["Order in which pet units appear"],
				usage = L["Order in which player units appear"],
				order = 3,
				get = function()
					return GridLayoutPlus.db.profile.petOrder
				end,
				set = function(v)
					GridLayoutPlus.db.profile.petOrder = v
					updateNeeded = true
					GridLayoutPlus:RosterUpdate()
				end,
			},
			["roleOrder"] = {
				type = "text",
				name = L["Blizzard-tank order"],
				desc = L["Order in which blizzard tanks appear"],
				usage = L["Order in which player units appear"],
				order = 4,
				get = function()
					return GridLayoutPlus.db.profile.roleOrder
				end,
				set = function(v)
					GridLayoutPlus.db.profile.roleOrder = v
					updateNeeded = true
					GridLayoutPlus:RosterUpdate()
				end,
			},
			["maxGridColumns"] = {
				type = "range",
				name = L["Max number of Grid columns"],
				desc = L["Maximum number of Grid columns when auto-adjusting before hiding tanks/pets"],
				order = 5,
				min = 5,
				max = 18,
				step = 1,
				get = function()
					return GridLayoutPlus.db.profile.maxGridColumns
				end,
				set = function(v)
					GridLayoutPlus.db.profile.maxGridColumns = v
					updateNeeded = true
					GridLayoutPlus:UpdateLayout()
				end,
			},
			["warnPetsHidden"] = {
				type = "toggle",
				name = L["Warn if pets are hidden"],
				desc = L["Show warning when pet units are forced to be hidden"],
				order = 6,
				get = function()
					return GridLayoutPlus.db.profile.warnPetsHidden
				end,
				set = function(v)
					GridLayoutPlus.db.profile.warnPetsHidden = v
					updateNeeded = true
					GridLayoutPlus:UpdateLayout()
				end,
			},
			["warnTanksHidden"] = {
				type = "toggle",
				name = L["Warn if tanks are hidden"],
				desc = L["Show warning when tank units are forced to be hidden"],
				order = 7,
				get = function()
					return GridLayoutPlus.db.profile.warnTanksHidden
				end,
				set = function(v)
					GridLayoutPlus.db.profile.warnTanksHidden = v
					updateNeeded = true
					GridLayoutPlus:UpdateLayout()
				end,
			},
		},
	}
	self:RegisterEvent("Grid_Enabled", "FullUpdate")
	self:RegisterEvent("Grid_JoinedRaid", "FullUpdate")
	self:RegisterEvent("Grid_JoinedBattleground", "FullUpdate")
	self:RegisterEvent("Grid_UnitJoined", "RosterUpdate")
	self:RegisterEvent("Grid_UnitLeft", "RosterUpdate")
	self:RegisterEvent("Grid_UnitChanged", "UnitUpdate")
	self:RegisterEvent("Grid_LeftParty", "ClearLocals")
	self:RegisterEvent("Grid_Disabled", "ClearLocals")
	if oRA then
		self:RegisterEvent("oRA_MainTankUpdate", "RATanksUpdate")
		self:RegisterEvent("oRA_JoinedRaid", "RATanksUpdate")
	elseif CT_RA_MainTanks then
		hooksecurefunc("CT_RAOptions_UpdateMTs", function() self:RATanksUpdate() end)
	end
	self:RegisterEvent("GridLayoutPlus_UpdateLayout", "UpdateLayout")
	self:ScheduleRepeatingEvent("GridLayoutPlus_UpdateLayout", 3)
	self:FullUpdate()
end


-- Event handlers

function GridLayoutPlus:FullUpdate()
	self:RATanksUpdate()
	self:RosterUpdate()
	updateNeeded = true
	self:UpdateLayout()
end

function GridLayoutPlus:UnitUpdate(name, unitid)
	local u = RL:GetUnitObjectFromName(name)
	if u and (u.class ~= "PET") then
		if (GridLayoutPlus.db.profile.layoutSorting == "By Group")
				or (GridLayoutPlus.db.profile.showTanks and (((tankType ~= "RA") and ((u.role == "MAINTANK") or (u.role == "MAINASSIST")))
														or ((tankType == "Blizz") and (u.role ~= "MAINTANK") and (u.role ~= "MAINASSIST")))) then
			self:RosterUpdate()
		end
	end
end

function GridLayoutPlus:RATanksUpdate()
	local ratanktable = nil
	local numberoftanks = 0
	if GridLayoutPlus.db.profile.showTanks and oRA and oRA.maintanktable then
		ratanktable = oRA.maintanktable
	elseif GridLayoutPlus.db.profile.showTanks and CT_RA_MainTanks then
		ratanktable = CT_RA_MainTanks
	end
	if ratanktable then
		for key, val in pairs(ratanktable) do
			local unitid = RL:GetUnitIDFromName(val)
			if unitid and UnitExists(unitid) and UnitPlayerOrPetInRaid(unitid) then
				numberoftanks = numberoftanks + 1
			end
		end
	end
	tankColumns = floor((numberoftanks / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))
	if (tankColumns > 0) then
		tankType = "RA"
		if oRA and oRA.maintanktable then
			raTankList = table.concat(oRA.maintanktable, ',')
		elseif CT_RA_MainTanks then
			raTankList = table.concat(CT_RA_MainTanks, ',')
		end
	elseif (tankType == "RA") then
		tankType = nil
		raTankList = nil
	end
	updateNeeded = true
end

function GridLayoutPlus:RosterUpdate()
	local maxsubgroup = 0
	local numberofpotentialpets = 0
	local numberoftanks = 0
	for u in RL:IterateRoster(nil) do
		if (GridLayoutPlus.db.profile.layoutSorting == "By Group") and (u.subgroup > maxsubgroup) then
			maxsubgroup = u.subgroup
		end
		if GridLayoutPlus.db.profile.showPets and ((u.class == "HUNTER") or (u.class == "WARLOCK")) then
			numberofpotentialpets = numberofpotentialpets + 1
		end
		if GridLayoutPlus.db.profile.showTanks and (tankType ~= "RA") and ((u.role == "MAINTANK") or (u.role == "MAINASSIST")) then
			numberoftanks = numberoftanks + 1
		end
	end
	if (GridLayoutPlus.db.profile.layoutSorting == "By Group") and (playerColumns ~= floor((maxsubgroup * 5 / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))) then
		playerColumns = floor((maxsubgroup * 5 / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))
		updateNeeded = true
	elseif (playerColumns ~= floor((GetNumRaidMembers() / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))) then
		playerColumns = floor((GetNumRaidMembers() / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))
		updateNeeded = true
	end
	if (petColumns ~= floor((numberofpotentialpets / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))) then
		petColumns = floor((numberofpotentialpets / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))
		updateNeeded = true
	end
	if (tankColumns ~= floor((numberoftanks / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))) then
		tankColumns = floor((numberoftanks / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))
		if (tankColumns > 0) then
			tankType = "Blizz"
		elseif (tankType == "Blizz") then
			tankType = nil
		end
		updateNeeded = true
	end
end

function GridLayoutPlus:ClearLocals()
	playerColumns = 0
	petColumns = 0
	tankColumns = 0
	tankType = nil
	raTankList = nil
end


-- Update layout

function GridLayoutPlus:UpdateLayout()
	if (updateNeeded) then
		if GridLayoutPlus.db.profile.autoLayout and (GetNumRaidMembers() > 0) and (not InCombatLockdown()) then
			local canfitpets = nil
			local canfittanks = nil
			local groupindex = 1
			-- Can fit pets / tanks
			if (GridLayoutPlus.db.profile.maxGridColumns >= (playerColumns + petColumns + tankColumns)) then
				canfitpets = true
				canfittanks = true
			elseif (GridLayoutPlus.db.profile.maxGridColumns >= (playerColumns + petColumns)) then
				canfitpets = true
				canfittanks = nil
			else
				canfitpets = nil
				canfittanks = nil
			end
			-- Alter layout
			if GridLayoutPlus.db.profile.showTanks and canfittanks and (tankType == "RA") then
				GridLayoutPlus.DefaultLayout[groupindex] = {
					nameList = raTankList,
					sortMethod = "INDEX",
					unitsPerColumn = GridLayoutPlus.db.profile.unitsPerColumn,
				}
				groupindex = groupindex + 1
			elseif GridLayoutPlus.db.profile.showTanks and canfittanks and (tankType == "Blizz") then
				GridLayoutPlus.DefaultLayout[groupindex] = {
					groupFilter = GridLayoutPlus.db.profile.roleOrder,
					groupingOrder = GridLayoutPlus.db.profile.roleOrder,
					groupBy = "ROLE",
					unitsPerColumn = GridLayoutPlus.db.profile.unitsPerColumn,
				}
				groupindex = groupindex + 1
			end
			if GridLayoutPlus.db.profile.showTanks and canfittanks and GridLayoutPlus.db.profile.showTankSpacer then
				GridLayoutPlus.DefaultLayout[groupindex] = {
					groupFilter = ""
				}
				groupindex = groupindex + 1
			end
			if (GridLayoutPlus.db.profile.layoutSorting == "By Group") then
				local groupnumber = 1
				local strgroupfilter
				for i = 1, playerColumns do
					strgroupfilter = ""
					for j = groupnumber,  (groupnumber + (GridLayoutPlus.db.profile.unitsPerColumn / 5) - 1) do
						strgroupfilter = strgroupfilter .. j .. ","
						groupnumber = j
					end
					GridLayoutPlus.DefaultLayout[groupindex] = {
						groupFilter = strgroupfilter,
						groupingOrder = GridLayoutPlus.db.profile.classOrder,
						groupBy = "CLASS",
						unitsPerColumn = GridLayoutPlus.db.profile.unitsPerColumn,
					}
					groupnumber = groupnumber + 1
					groupindex = groupindex + 1
				end
			elseif (GridLayoutPlus.db.profile.layoutSorting == "By Class") then
				GridLayoutPlus.DefaultLayout[groupindex] = {
					groupFilter = GridLayoutPlus.db.profile.classOrder,
					groupingOrder = GridLayoutPlus.db.profile.classOrder,
					groupBy = "CLASS",
					unitsPerColumn = GridLayoutPlus.db.profile.unitsPerColumn,
				}
				groupindex = groupindex + 1
			end
			if GridLayoutPlus.db.profile.showPets and canfitpets and GridLayoutPlus.db.profile.showPetSpacer then
				GridLayoutPlus.DefaultLayout[groupindex] = {
					groupFilter = ""
				}
				groupindex = groupindex + 1
			end
			if GridLayoutPlus.db.profile.showPets and canfitpets and (petColumns > 0) then
				GridLayoutPlus.DefaultLayout[groupindex] = {
					isPetGroup = true,
					groupFilter = GridLayoutPlus.db.profile.petOrder,
					groupingOrder = GridLayoutPlus.db.profile.petOrder,
					groupBy = "CLASS",
					filterOnPet = true,
					unitsPerColumn = GridLayoutPlus.db.profile.unitsPerColumn,
				}
				groupindex = groupindex + 1
			end
			for i = groupindex, 10 do
				GridLayoutPlus.DefaultLayout[i] = nil
			end
			-- Refresh layout
			if GridLayout.db.profile.layout ~= "GridLayoutPlus" then
				GridLayout.db.profile.layout = "GridLayoutPlus"
				GridLayout:LoadLayout("GridLayoutPlus")
			else
				GridLayout:ReloadLayout()
			end
			-- Warn if hidden units
			if GridLayoutPlus.db.profile.warnPetsHidden and GridLayoutPlus.db.profile.showPets and (petColumns > 0) and (not canfitpets) then
				self:Print(L["Your maxGridColumns setting is causing pet units to be hidden"])
			end
			if GridLayoutPlus.db.profile.warnTanksHidden and GridLayoutPlus.db.profile.showTanks and (tankColumns > 0) and (not canfittanks) then
				self:Print(L["Your maxGridColumns setting is causing tank units to be hidden"])
			end
			updateNeeded = nil
		elseif GridLayoutPlus.db.profile.autoLayout and (GetNumRaidMembers() > 0) and InCombatLockdown() then
			self:ScheduleLeaveCombatAction(self, "RosterUpdate")
		end
	end
end
